import dataclasses
import typing
import PIL.Image
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Make a transform function for image and prompt list.

    The interface of the transform function is:
    - Input: (PIL.Image.Image, torch.Tensor), where the tensor is a scalar index of the prompt list.
    - Output: (str, [torch.Tensor]), where the string is the prompt, and the list of tensors is the image.
    """

    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        image_transform: typing.Callable
        prompts: typing.List[str]

    @dataclasses.dataclass
    class Outputs:
        transform: typing.Callable

    def execute(self, inputs):
        return self.Outputs(PromptListImageTransform(inputs.image_transform, inputs.prompts))

    def dry_run(self, inputs):
        return self.execute(inputs)


class PromptListImageTransform:
    def __init__(self, image_transform, prompts):
        self._image_transform = image_transform
        self._prompts = prompts

    def __call__(self, inputs, targets):
        assert isinstance(inputs, PIL.Image.Image)
        image_tensor = self._image_transform(inputs)
        assert isinstance(image_tensor, torch.Tensor)
        assert isinstance(targets, torch.Tensor) and targets.dim() == 0, f"targets: {targets}"
        prompt = self._prompts[targets.item()]
        return (prompt, [image_tensor]), targets



import dataclasses
import typing
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Map classification predictions back to detection predictions or targets.

    This tasks is to apply the second-stage IC model predictions back to the OD data. It takes IC prediction (images cropped from OD dataset), OD predictions or targets,
    and the index mapping as input, outputs detection predictions with scores, where the score can be IC score, OD score, product or average.

    The detection prediction classes will always be replaced by classification.

    Inputs:
        classifications: tensor of shape (N, C) where C is num of classes, with the predicted probabilities; or (N,) with the predicted classes only
        detections: a list of tensors with (N_i, 6) with scores or (N_i, 5) without scores, where N_i is num of boxes for image i
        index_mappings: IC image index to a tuple (original image index in OD dataset, box index in that image)

    Outputs:
        detection_predictions: a list of tensors with (N_i, 6) with scores, where N_i is num of boxes for image i
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        classifications: torch.Tensor
        detections: typing.List[torch.Tensor]
        index_mappings: typing.List[typing.Tuple[int, int]]

    @dataclasses.dataclass
    class Config:
        score_type: typing.Literal['ic', 'od', 'prod', 'avg'] = 'ic'

    @dataclasses.dataclass
    class Outputs:
        detection_predictions: typing.List[torch.Tensor]

    def execute(self, inputs):
        detections = [torch.as_tensor(d) for d in inputs.detections]
        if inputs.classifications.shape[0] != len(inputs.index_mappings):
            raise ValueError(f"IC predictions have different samples from the IC to OD index mapping: {inputs.classifications.shape[0]} vs {len(inputs.index_mappings)}")
        if detections[0].shape[1] == 5 and (self.config.score_type != 'ic' or inputs.classifications.ndim == 1):
            raise ValueError("When OD scores not provided, score_type has to be ic and IC scores have to be provided!")
        if inputs.classifications.ndim == 1 and (self.config.score_type != 'od' or detections[0].shape[1] == 5):
            raise ValueError("When IC scores not provided, score_type has to be od and OD scores have to be provided!")
        if self.config.score_type not in ['ic', 'od', 'prod', 'avg']:
            raise ValueError(f"score type must be one of 'ic', 'od', 'prod', 'avg', got {self.config.score_type}")

        if inputs.classifications.ndim == 2:
            ic_scores, ic_pred_classes = torch.max(inputs.classifications, dim=-1)
        else:  # if no IC scores are provided
            ic_pred_classes = inputs.classifications
            ic_scores = torch.ones_like(ic_pred_classes, dtype=float)

        output = [torch.zeros((img_detections.shape[0], 6)) for img_detections in detections]
        for ic_pred_class, ic_score, (img_id, box_id_in_img) in zip(ic_pred_classes, ic_scores, inputs.index_mappings):
            if self.config.score_type == 'ic':
                score = ic_score
            else:
                od_score = detections[img_id][box_id_in_img][1]
                if self.config.score_type == 'od':
                    score = od_score
                elif self.config.score_type == 'prod':
                    score = ic_score * od_score
                else:
                    score = (ic_score + od_score) * 0.5
            output[img_id][box_id_in_img][:2] = torch.tensor([ic_pred_class, score])
            output[img_id][box_id_in_img][2:] = detections[img_id][box_id_in_img][-4:]

        return self.Outputs(output)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import unittest.mock
import PIL.Image
import torch
from irisml.tasks.make_classification_dataset_from_object_detection import Task
from utils import FakeDataset


class FakeDatasetWithGetTargets(torch.utils.data.Dataset):
    def __init__(self, targets):
        self._targets = targets
        self.reset()

    def __len__(self):
        return len(self._targets)

    def __getitem__(self, index):
        self._getitem_args.append(index)
        return PIL.Image.new('RGB', (8, 8)), self._targets[index]

    def get_targets(self, index):
        self._get_targets_args.append(index)
        return self._targets[index]

    def reset(self):
        self._getitem_args = []
        self._get_targets_args = []


class TestMakeClassificationDatasetFromObjectDetection(unittest.TestCase):
    def test_simple(self):
        fake_image = PIL.Image.new('RGB', (32, 32))
        dataset = FakeDataset([(fake_image, torch.tensor([[0, 0, 0, 0.5, 0.5], [1, 0, 0, 0.1, 0.1]])),
                               (fake_image, torch.zeros(0, 5)),
                               (fake_image, torch.tensor([[2, 0, 0, 0.5, 0.5], [3, 0, 0, 0.1, 0.1]])),
                               (fake_image, torch.zeros(0, 5))])

        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 4)
        self.assertEqual(outputs.dataset[0][0].size, (16, 16))
        self.assertEqual(outputs.dataset[0][1], 0)
        self.assertEqual(outputs.dataset[1][0].size, (3, 3))
        self.assertEqual(outputs.dataset[1][1], 1)
        self.assertEqual(outputs.dataset[2][0].size, (16, 16))
        self.assertEqual(outputs.dataset[2][1], 2)
        self.assertEqual(outputs.dataset[3][0].size, (3, 3))
        self.assertEqual(outputs.dataset[3][1], 3)
        self.assertEqual(outputs.index_mappings, [(0, 0), (0, 1), (2, 0), (2, 1)])

    def test_given_multiclass_dataset(self):
        fake_image = PIL.Image.new('RGB', (32, 32))
        dataset = FakeDataset([(fake_image, torch.tensor(0)),
                               (fake_image, torch.tensor(1))])
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 2)
        self.assertEqual(outputs.dataset[0][0].size, (32, 32))
        self.assertEqual(outputs.dataset[0][1], 0)
        self.assertEqual(outputs.dataset[1][1], 1)
        self.assertIsNone(outputs.index_mappings)

    def test_given_multilabel_dataset(self):
        fake_image = PIL.Image.new('RGB', (32, 32))
        dataset = FakeDataset([(fake_image, torch.tensor([0, 1])),
                               (fake_image, torch.tensor([2, 3, 4]))])
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 2)
        self.assertEqual(outputs.dataset[0][0].size, (32, 32))
        self.assertEqual(outputs.dataset[0][1].dim(), 1)
        self.assertEqual(outputs.dataset[1][1][2], 4)
        self.assertIsNone(outputs.index_mappings)

    def test_get_targets(self):
        dataset = FakeDatasetWithGetTargets([torch.tensor([[0, 0, 0, 0.5, 0.5], [1, 0, 0, 0.1, 0.1]]), torch.tensor([[0, 0, 0, 1, 1]]), torch.empty(0, 5)])

        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertEqual(dataset._getitem_args, [0])
        self.assertEqual(dataset._get_targets_args, [0, 1, 2])
        self.assertEqual(len(outputs.dataset), 3)

        dataset.reset()
        self.assertEqual(outputs.dataset.get_targets(0), torch.tensor(0))
        self.assertEqual(outputs.dataset.get_targets(1), torch.tensor(1))
        self.assertEqual(outputs.dataset.get_targets(2), torch.tensor(0))
        self.assertEqual(dataset._getitem_args, [])
        self.assertEqual(dataset._get_targets_args, [0, 0, 1])

        dataset.reset()
        self.assertEqual(outputs.dataset[0][1], torch.tensor(0))
        self.assertEqual(outputs.dataset[1][1], torch.tensor(1))
        self.assertEqual(outputs.dataset[2][1], torch.tensor(0))
        self.assertEqual(dataset._getitem_args, [0, 0, 1])
        self.assertEqual(dataset._get_targets_args, [])



import unittest
from irisml.tasks.create_classification_prompt_generator import Task


class TestCreateClassificationPromptGenerator(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config()).execute(Task.Inputs())

        results = outputs.generator('random_string')
        self.assertGreater(len(results), 1)
        self.assertTrue(all('random_string' in r for r in results))

    def test_template(self):
        outputs = Task(Task.Config(name='template', template='This is an image of {}')).execute(Task.Inputs())

        self.assertEqual(outputs.generator('label_name'), ['This is an image of label_name'])

        # No template
        outputs = Task(Task.Config(name='template')).execute(Task.Inputs())

        self.assertEqual(outputs.generator('label_name'), ['label_name'])

        # Without {} in the template
        with self.assertRaises(ValueError):
            Task(Task.Config(name='template', template='This is an image of')).execute(Task.Inputs())



import unittest
import PIL.Image
import torch
import torchvision
from irisml.tasks.make_prompt_list_image_transform import Task


class TestMakePromptListImageTransform(unittest.TestCase):
    def test_simple(self):
        image_transform = torchvision.transforms.ToTensor()
        prompts = ['prompt1', 'prompt2']
        outputs = Task(Task.Config()).execute(Task.Inputs(image_transform, prompts))
        transform = outputs.transform

        self.assertEqual(transform(PIL.Image.new('RGB', (1, 1)), torch.tensor(0))[0][0], 'prompt1')
        self.assertEqual(transform(PIL.Image.new('RGB', (1, 1)), torch.tensor(1))[0][0], 'prompt2')
        self.assertTrue(torch.equal(transform(PIL.Image.new('RGB', (1, 1)), torch.tensor(0))[0][1][0], torch.zeros(3, 1, 1)))
        self.assertTrue(torch.equal(transform(PIL.Image.new('RGB', (1, 1)), torch.tensor(0))[0][1][0], torch.zeros(3, 1, 1)))
        self.assertTrue(torch.equal(transform(PIL.Image.new('RGB', (1, 1)), torch.tensor(1))[1], torch.tensor(1)))



import collections
import logging
import re

logger = logging.getLogger(__name__)


def build_optimizer(name: str, parameters, base_lr, weight_decay, momentum):
    import torch  # Due to the `import torch.distributed.optim` statements below, this import must be here.
    if name == 'sgd':
        return torch.optim.SGD(parameters, lr=base_lr, weight_decay=weight_decay, momentum=momentum)
    elif name == 'adam':
        return torch.optim.Adam(parameters, lr=base_lr, weight_decay=weight_decay)
    elif name == 'amsgrad':
        return torch.optim.Adam(parameters, lr=base_lr, weight_decay=weight_decay, amsgrad=True)
    elif name == 'adamw':
        return torch.optim.AdamW(parameters, lr=base_lr, weight_decay=weight_decay)
    elif name == 'adamw_amsgrad':
        return torch.optim.AdamW(parameters, lr=base_lr, weight_decay=weight_decay, amsgrad=True)
    elif name == 'rmsprop':
        return torch.optim.RMSprop(parameters, lr=base_lr, weight_decay=weight_decay, momentum=momentum)
    elif name == 'zero_adam':
        import torch.distributed.optim  # Load lazily since importing this module will create a temporary directory.
        return torch.distributed.optim.ZeroRedundancyOptimizer(parameters, optimizer_class=torch.optim.Adam, lr=base_lr, weight_decay=weight_decay)
    elif name == 'zero_amsgrad':
        import torch.distributed.optim  # Load lazily since importing this module will create a temporary directory.
        return torch.distributed.optim.ZeroRedundancyOptimizer(parameters, optimizer_class=torch.optim.Adam, lr=base_lr, weight_decay=weight_decay, amsgrad=True)
    elif name == 'zero_adamw':
        import torch.distributed.optim  # Load lazily since importing this module will create a temporary directory.
        return torch.distributed.optim.ZeroRedundancyOptimizer(parameters, optimizer_class=torch.optim.AdamW, lr=base_lr, weight_decay=weight_decay)
    elif name == 'zero_adamw_amsgrad':
        import torch.distributed.optim  # Load lazily since importing this module will create a temporary directory.
        return torch.distributed.optim.ZeroRedundancyOptimizer(parameters, optimizer_class=torch.optim.AdamW, lr=base_lr, weight_decay=weight_decay, amsgrad=True)
    else:
        raise ValueError(f"Unsupported optimizer: {name}")


class OptimizerFactory:
    def __init__(self, name: str, base_lr, weight_decay=0, momentum=0, no_weight_decay_param_names=None, no_weight_decay_module_class_names=None, lr_scale_param_names=None):
        self._name = name
        self._base_lr = base_lr
        self._weight_decay = weight_decay
        self._momentum = momentum
        self._no_weight_decay_param_name_patterns = [re.compile(p) for p in (no_weight_decay_param_names or [])]
        self._no_weight_decay_module_class_name_patterns = [re.compile(p) for p in (no_weight_decay_module_class_names or [])]
        self._lr_scale_param_name_patterns = [(re.compile(k), v) for k, v in (lr_scale_param_names or [])]

    def __call__(self, model):
        params = []
        no_weight_decay_params = set()

        for name, module in model.named_modules():
            if any(p.match(type(module).__name__) for p in self._no_weight_decay_module_class_name_patterns):
                logger.debug(f"Disabling weight_decay for module {name}")
                no_weight_decay_params.update(module.parameters())

        param_groups_dict = collections.defaultdict(list)
        for name, param in model.named_parameters():
            weight_decay = None
            lr = None

            if param in no_weight_decay_params:
                weight_decay = 0.0

            if weight_decay is None and any(p.match(name) for p in self._no_weight_decay_param_name_patterns):
                logger.debug(f"Disabling weight_decay for param {name}")
                weight_decay = 0.0

            lr_scale_entry = next((p for p in self._lr_scale_param_name_patterns if p[0].match(name)), None)
            if lr_scale_entry:
                logger.debug(f"Scaling lr for param {name} with {lr_scale_entry[1]}")
                lr = self._base_lr * lr_scale_entry[1]

            param_groups_dict[(weight_decay, lr)].append(param)

        assert len(list(model.parameters())) == sum(len(p) for p in param_groups_dict.values())

        # Construct param_groups parameter.
        param_groups = []
        for (weight_decay, lr), params in param_groups_dict.items():
            group = {'params': params}
            if weight_decay is not None:
                group['weight_decay'] = weight_decay
            if lr is not None:
                group['lr'] = lr
            param_groups.append(group)

        return build_optimizer(self._name, param_groups, self._base_lr, self._weight_decay, self._momentum)



import inspect
import typing
import PIL.Image
import torch.multiprocessing
import torch.utils.data


class Dataset(torch.utils.data.Dataset):
    """Dataset wrapper to support both image-only transforms and image-and-targets transforms."""
    def __init__(self, dataset, transform):
        self._dataset = dataset
        self._transform = transform
        if transform is not None:
            self._num_transform_args = self._get_num_args(transform)
            assert 1 <= self._num_transform_args <= 2
        else:
            self._num_transform_args = 0

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, index):
        item = self._dataset[index]
        assert len(item) == 2
        if self._num_transform_args == 1:
            return self._transform(item[0]), item[1]
        elif self._num_transform_args == 2:
            return self._transform(*item)
        elif self._num_transform_args == 0:
            return item

    @staticmethod
    def _get_num_args(function):
        args = inspect.getfullargspec(function).args
        if args[0] == 'self':
            return len(args) - 1
        return len(args)


def collate(batch) -> typing.List:
    """Handle batching.

    Supported batch formats are:
        [(Tensor[C, H, W], int), ...]  # multiclass classification
        [(Tensor[C, H, W], [int, ...]), ...]  # multiclass or multilabel classification
        [(Tensor[C, H, W], Tensor[T]), ...]  # multiclass or multilabel classification
        [(Tensor[C, H, W], [float, ...]), ...]  # regression
        [(Tensor[C, H, W], [[tag, x, y, x2, y2], ...]), ...]  # object detection
        [(Tensor[C, H, W], Tensor[T, 5]), ...]  # object detection
        [Tensor[C, H, W], ...]  # prediction
        [((str, [Tensor[C, H, W], ...]), *), ...]  # text generation
    """

    if isinstance(batch[0], tuple):
        if isinstance(batch[0][0], tuple) and isinstance(batch[0][0][0], str):
            return [[b[0] for b in batch], _collate_targets([b[1] for b in batch])]
        else:
            return [_collate_tensor([b[0] for b in batch]), _collate_targets([b[1] for b in batch])]
    elif isinstance(batch[0], torch.Tensor):
        return [_collate_tensor(batch), None]
    elif isinstance(batch[0], str):
        return batch

    raise TypeError(f"Unexpected type: {type(batch[0])}")


def _collate_tensor(batch: typing.List):
    """
    Expected input types are:
    - List[Tensor[C, H, W]]
    - List[Tuple[Tensor[C, H, W], Tensor]]
    """
    if isinstance(batch[0], torch.Tensor):
        if all(b.shape == (1,) for b in batch):
            return torch.cat(batch)

        if all(batch[0].shape == b.shape for b in batch):
            return torch.stack(batch, 0)

        return batch
    elif isinstance(batch[0], tuple):
        # For image-text dataset.
        return _collate_tensor([b[0] for b in batch]), _collate_tensor([b[1] for b in batch])
    elif isinstance(batch[0], PIL.Image.Image):
        return batch

    raise TypeError(f"Unexpected type: {type(batch[0])}")


def _collate_targets(batch: typing.List):
    if all(b is None for b in batch):
        return batch

    if isinstance(batch[0], torch.Tensor):
        return _collate_tensor(batch)

    if isinstance(batch[0], list):
        if any(b and isinstance(b[0], (list, tuple)) for b in batch):  # batch = [[[...], ...], ...]
            return [torch.tensor(b).reshape(-1, 5) for b in batch]
        elif any(b and isinstance(b[0], (int, float)) for b in batch):  # batch = [[int, ...], ...]
            return _collate_tensor([torch.tensor(b) for b in batch])
        elif any(x and all(isinstance(y, str) for y in x) for x in batch):  # caption
            return batch
        elif any(x and all(len(y) == 2 and isinstance(y[0], str) and isinstance(y[1], int) for y in x) for x in batch):  # image-text matching
            return batch
        elif not any(batch):
            return batch
    elif isinstance(batch[0], int):
        return torch.tensor(batch)
    elif isinstance(batch[0], str):
        return batch

    raise TypeError(f"Unexpected target type: {batch}")


def build_dataloader(dataset, transform, batch_size: int, num_processes: int = 1, num_workers: int = 4, shuffle=True, drop_last=True):
    """Build a DataLoader class.

    If the dataset is smaller than batch_size*num_processes, it will oversample the dataset so that there is at least one batch per epoch.
    """
    my_dataset = Dataset(dataset, transform)
    sampler = (torch.utils.data.RandomSampler(my_dataset, replacement=(batch_size * num_processes) > len(dataset), num_samples=max(batch_size * num_processes, len(dataset))) if shuffle
               else torch.utils.data.SequentialSampler(my_dataset))
    return torch.utils.data.DataLoader(my_dataset, batch_size=batch_size, num_workers=num_workers, sampler=sampler, drop_last=drop_last,
                                       pin_memory=True, collate_fn=collate, multiprocessing_context=torch.multiprocessing.get_context('fork'))



import unittest
import PIL.Image
import torch
import torchvision.transforms
from irisml.tasks.make_vqa_image_transform import Task


class TestMakeVqaImageTransform(unittest.TestCase):
    def test_simple(self):
        image_transform = torchvision.transforms.ToTensor()

        outputs = Task(Task.Config()).execute(Task.Inputs(image_transform))
        transform = outputs.transform

        transform_outputs = transform((PIL.Image.new('RGB', (32, 32)), 'What is this?'), 'R2D2')
        self.assertEqual(transform_outputs[0][0], '<|image|> question: What is this? answer:')
        self.assertIsInstance(transform_outputs[0][1], torch.Tensor)
        self.assertEqual(transform_outputs[1], 'R2D2')



import unittest
import PIL.Image
import torch.utils.data
from irisml.tasks.get_questions_from_vqa_dataset import Task


class FakeDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        self._data = data

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)


class TestGetQuestionsFromVqaDataset(unittest.TestCase):
    def test_simple(self):
        data = [((PIL.Image.new('RGB', (224, 224)), 'What is this?'), 'Apple'),
                ((PIL.Image.new('RGB', (224, 224)), 'Where is it?'), 'Paris')]

        dataset = FakeDataset(data)

        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertEqual(outputs.questions, ['What is this?', 'Where is it?'])



import unittest
import unittest.mock
import PIL.Image
import torch
from irisml.tasks.make_detection_dataset_from_predictions import Task
from utils import FakeDataset


class TestMakeClassificationDatasetFromObjectDetection(unittest.TestCase):
    def build_dataset(self):
        fake_image = PIL.Image.new('RGB', (10, 10))
        dataset = FakeDataset([(fake_image, torch.tensor([[0, 0, 0, 0.5, 0.5], [1, 0, 0, 0.1, 0.1]])),
                               (fake_image, torch.zeros(0, 5)),
                               (fake_image, torch.tensor([[2, 0, 0, 0.5, 0.5], [3, 0, 0, 0.1, 0.1]]))])
        return dataset

    def test_simple(self):
        dataset = self.build_dataset()

        predictions_xyxy = [torch.tensor([[0, 0.1, 0.1, 0.4, 0.5], [1, 0.1, 0.1, 0.2, 0.2]]), torch.empty(0, 5), torch.tensor([[2, 0, 0, 1, 1]])]
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset, predictions_xyxy))

        self.assertEqual(len(outputs.dataset), 3)
        for im, _ in outputs.dataset:
            self.assertEqual(im.size, (10, 10))

        # Test normalized coordinates.
        self.assertTrue(torch.equal(outputs.dataset[0][1], torch.tensor([[0, 0.1, 0.1, 0.4, 0.5], [1, 0.1, 0.1, 0.2, 0.2]])))
        self.assertTrue(torch.equal(outputs.dataset[1][1], torch.empty(0, 5)))
        self.assertTrue(torch.equal(outputs.dataset[2][1], torch.tensor([[2, 0, 0, 1, 1]])))

    def test_normalize(self):
        dataset = self.build_dataset()
        predictions_absolute = [torch.tensor([[0, 1, 1, 4, 5], [1, 1, 1, 2, 2]]), torch.empty(0, 5), torch.tensor([[2, 0, 0, 10, 10]])]
        outputs = Task(Task.Config(normalized=False)).execute(Task.Inputs(dataset, predictions_absolute))
        self.assertTrue(torch.equal(outputs.dataset[0][1], torch.tensor([[0, 0.1, 0.1, 0.4, 0.5], [1, 0.1, 0.1, 0.2, 0.2]])))
        self.assertTrue(torch.equal(outputs.dataset[1][1], torch.empty(0, 5)))
        self.assertTrue(torch.equal(outputs.dataset[2][1], torch.tensor([[2, 0, 0, 1, 1]])))

    def test_invalid_coordinates(self):
        dataset = self.build_dataset()
        # Some coordindates are not within [0, 1].
        predictions_xyxy_invalid = [torch.tensor([[0, 0.1, 0.1, 1.1, 0.5], [1, 0.1, 0.1, 0.2, 0.2]]), torch.empty(0, 5), torch.tensor([[2, -0.2, 0, 1.5, 1.2]])]
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset, predictions_xyxy_invalid))
        self.assertTrue(torch.equal(outputs.dataset[0][1], torch.tensor([[0, 0.1, 0.1, 1.0, 0.5], [1, 0.1, 0.1, 0.2, 0.2]])))
        self.assertTrue(torch.equal(outputs.dataset[1][1], torch.empty(0, 5)))
        self.assertTrue(torch.equal(outputs.dataset[2][1], torch.tensor([[2, 0, 0, 1, 1]])))

    def test_predictions_with_score(self):
        dataset = self.build_dataset()
        predictions_with_scores = [torch.tensor([[0, 0.8, 0.1, 0.1, 0.9, 0.5], [1, 0.3, 0.1, 0.1, 0.2, 0.2]]), torch.empty(0, 6), torch.tensor([[2, 0.2, 0.2, 0, 1.5, 1.2]])]
        outputs = Task(Task.Config(score_threshold=0.5)).execute(Task.Inputs(dataset, predictions_with_scores))
        self.assertTrue(torch.equal(outputs.dataset[0][1], torch.tensor([[0, 0.1, 0.1, 0.9, 0.5]])))
        self.assertTrue(torch.equal(outputs.dataset[1][1], torch.empty(0, 5)))
        self.assertTrue(torch.equal(outputs.dataset[2][1], torch.empty(0, 5)))

    def test_predictions_with_score_xywh(self):
        dataset = self.build_dataset()
        predictions_with_scores = [torch.tensor([[0, 0.8, 0.1, 0.1, 0.5, 0.6], [1, 0.3, 0.1, 0.1, 0.3, 0.3]]), torch.empty(0, 6), torch.tensor([[2, 0.2, 0, 0, 1, 1]])]
        outputs = Task(Task.Config(score_threshold=0.5)).execute(Task.Inputs(dataset, predictions_with_scores))
        self.assertTrue(torch.equal(outputs.dataset[0][1], torch.tensor([[0, 0.1, 0.1, 0.5, 0.6]])))
        self.assertTrue(torch.equal(outputs.dataset[1][1], torch.empty(0, 5)))
        self.assertTrue(torch.equal(outputs.dataset[2][1], torch.empty(0, 5)))



import dataclasses
import logging
from typing import List, Optional
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Find incorrect classification indices for multilabel classification.

    Inputs:
        predictions_list (List[LongTensor]): List of predicted indices. Each element must be 1D.
        targets (Tensor): 1D tensor of target indices. Either targets or targets_list must be provided.
        targets_list (List[Tensor]): List of target indices. Each element must be 1D.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        # predictions: Optional[torch.Tensor] = None # NYI
        predictions_list: List[torch.Tensor]
        targets: Optional[torch.Tensor] = None
        targets_list: Optional[List[torch.Tensor]] = None

    @dataclasses.dataclass
    class Outputs:
        indices: torch.Tensor

    def execute(self, inputs):
        if inputs.targets is None and inputs.targets_list is None:
            raise ValueError("targets or targets_list must be provided")
        if inputs.targets is not None and inputs.targets_list is not None:
            raise ValueError("only one of targets or targets_list must be provided")

        if any(torch.is_floating_point(prediction) for prediction in inputs.predictions_list):
            raise ValueError("predictions_list must not be float.")

        predictions_list = inputs.predictions_list
        targets_list = inputs.targets_list
        if targets_list is None:
            targets_list = [t.reshape(1) for t in inputs.targets]

        if len(predictions_list) != len(targets_list):
            raise ValueError("The number of predictions and targets must be the same.")

        indices = []
        for i in range(len(predictions_list)):
            if set(predictions_list[i].tolist()) != set(targets_list[i].tolist()):
                logger.info(f"Index {i}: Predicted: {predictions_list[i].tolist()}, Actual: {targets_list[i].tolist()}")
                indices.append(i)

        return self.Outputs(torch.tensor(indices, dtype=torch.int64))

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import typing
import irisml.core


class Task(irisml.core.TaskBase):
    """Create a prompt generator that returns a list of prompts for a given label.

    Config:
        templates (List[str]): List of templates to use for the generator. Each template must contain a single {}.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        templates: typing.List[str]

    @dataclasses.dataclass
    class Outputs:
        generator: typing.Callable[[str], typing.List[str]]

    def execute(self, inputs):
        if not self.config.templates:
            raise ValueError('Templates must be provided.')
        return self.Outputs(PromptGenerator(self.config.templates))

    def dry_run(self, inputs):
        return self.execute(inputs)


class PromptGenerator:
    def __init__(self, templates):
        self._templates = templates

    def __call__(self, label_name: str) -> typing.List[str]:
        return [t.format(label_name) for t in self._templates]



import dataclasses
import logging
import typing
import torch
import PIL.Image
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Convert an object detection dataset into a classification dataset."""
    VERSION = '0.1.3'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        index_mappings: typing.Optional[typing.List[typing.Tuple[int, int]]] = None  # map cropped image index in IC to a tuple (original image index in OD dataset, box index in that image)

    def execute(self, inputs):
        if (len(inputs.dataset) == 0 or inputs.dataset[0][1].dim() <= 1):
            # Return input dataset if it is already a multiclass or multilabel classification dataset.
            logger.debug("Dataset is already classification. Skip the task.")
            return self.Outputs(inputs.dataset)
        else:
            od_as_ic_dataset = ClassificationFromDetectionDataset(inputs.dataset)
            return self.Outputs(od_as_ic_dataset, od_as_ic_dataset.index_mappings)

    def dry_run(self, inputs):
        return self.execute(inputs)


class ClassificationFromDetectionDataset(torch.utils.data.Dataset):
    def __init__(self, dataset):
        self._dataset = dataset
        num_boxes = [len(self._get_targets_from_original(i)) for i in range(len(dataset))]
        self._index_mappings = []
        for i, n in enumerate(num_boxes):
            self._index_mappings.extend([(i, j) for j in range(n)])

        logger.info(f"Created a classificaiton dataset with {len(self._index_mappings)} samples. The source dataset has {len(self._dataset)} samples.")
        assert len(self._index_mappings) == sum(num_boxes)

    def __len__(self):
        return len(self._index_mappings)

    def __getitem__(self, index):
        image_index, box_index = self._index_mappings[index]
        image, targets = self._dataset[image_index]
        assert isinstance(image, PIL.Image.Image)
        w, h = image.size
        box = targets[box_index]
        cropped = image.crop((int(w * box[1]), int(h * box[2]), int(w * box[3]), int(h * box[4])))
        return cropped, box[0].int()

    def get_targets(self, index):
        image_index, box_index = self._index_mappings[index]
        targets = self._get_targets_from_original(image_index)
        box = targets[box_index]
        return box[0].int()

    def _get_targets_from_original(self, index):
        return self._dataset.get_targets(index) if hasattr(self._dataset, 'get_targets') else self._dataset[index][1]

    @property
    def index_mappings(self):
        return self._index_mappings



import dataclasses
from typing import Callable, Tuple
import PIL.Image
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Creates a transform function for VQA task.

    The transform function expects two arguments: inputs and targets. inputs is a tuple of (image, question), where
    image is a PIL image and question is a string. targets is a string. The transform function returns ((prompt, image_tensor), targets),
    where prompt is a string and image_tensor is a torch tensor
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        image_transform: Callable[[PIL.Image.Image], torch.Tensor]

    @dataclasses.dataclass
    class Outputs:
        transform: Callable[[Tuple[PIL.Image.Image, str], str], Tuple[Tuple[str, torch.Tensor], str]]

    def execute(self, inputs):
        transform = VqaImageTransform(inputs.image_transform)
        return self.Outputs(transform=transform)

    def dry_run(self, inputs):
        return self.execute(inputs)


class VqaImageTransform:
    def __init__(self, image_transform):
        self._image_transform = image_transform

    def __call__(self, inputs, targets):
        image, question = inputs
        assert isinstance(image, PIL.Image.Image)
        image_tensor = self._image_transform(image)
        prompt = f'<|image|> question: {question} answer:'
        return (prompt, image_tensor), targets



import dataclasses
import json
import logging
import typing
import pathlib
import torch
import torch.utils.data
from torchvision.transforms import ToPILImage
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Export coco dataset from a given torch dataset. Support IC and OD only.

    For IC dataset, target is either 0 or 1 dim array; for OD, target is 2-dim array with each row being [label, x1, y1, x2, y2] normalized.
    The images are saved to [dirpath]/[image_directory_name]/[image_id].[image_extension].

    In the COCO file, the file_name for each image is relative to dirpath, bbox coordinates are [x, y, w, h] absolute.
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset
        class_names: typing.List[str]

    @dataclasses.dataclass
    class Config:
        task_type: typing.Literal['classification_multiclass', 'classification_multilabel', 'object_detection']
        dirpath: pathlib.Path = pathlib.Path('.')
        image_directory_name: str = 'images'
        json_file_name: str = 'images'
        image_extension: typing.Literal['png', 'jpg'] = 'png'

    def execute(self, inputs):
        def _convert_bbox(bbox: list, w: int, h: int):
            return [bbox[0] * w, bbox[1] * h, (bbox[2] - bbox[0]) * w, (bbox[3] - bbox[1]) * h]

        if (self.config.dirpath / f'{self.config.json_file_name}.json').exists():
            raise RuntimeError(f'{self.config.json_file_name}.json already exists in {self.config.dirpath}!')

        image_dir = self.config.dirpath / self.config.image_directory_name
        image_dir.mkdir(parents=True, exist_ok=True)
        dataset = inputs.dataset

        coco = {"images": [], "annotations": []}
        coco["categories"] = [{"id": i + 1, "name": inputs.class_names[i]} for i in range(len(inputs.class_names))]
        anno_idx = 1
        tensor_to_pil = ToPILImage()

        for image_id, (image, targets) in enumerate(dataset, 1):
            image_path = image_dir / f"{image_id}.{self.config.image_extension}"
            if isinstance(image, torch.Tensor):
                image = tensor_to_pil(image)
            image.save(image_path)
            logger.info(f"Index {image_id - 1} image saved to {str(image_path)}.")

            w, h = image.size
            image_info = {"id": image_id, "width": w, "height": h, "file_name": f"{self.config.image_directory_name}/{image_path.name}"}
            coco["images"].append(image_info)

            if not isinstance(targets, torch.Tensor):
                targets = torch.tensor(targets)

            if self.config.task_type == 'classification_multiclass':
                annos = [{"id": anno_idx, "category_id": int(targets if targets.dim() == 0 else targets[0]) + 1, "image_id": image_id}]
            elif self.config.task_type == 'classification_multilabel':
                annos = [{"id": anno_idx + i, "category_id": int(t) + 1, "image_id": image_id} for i, t in enumerate(targets)]
            else:
                annos = [{"id": anno_idx + i, "category_id": int(t[0]) + 1, "image_id": image_id, "bbox": _convert_bbox(t[1:].tolist(), w, h)} for i, t in enumerate(targets)]

            coco['annotations'].extend(annos)
            anno_idx += len(annos)

        (self.config.dirpath / f"{self.config.json_file_name}.json").write_text(json.dumps(coco, indent=2))
        return self.Outputs()



import unittest
import json
import tempfile
import pathlib
import PIL.Image
import torch
from irisml.tasks.export_coco_from_torch_dataset import Task


class FakeDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        self._data = data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        return self._data[index]


class TestExportCOCOJsonFromDataset(unittest.TestCase):
    def test_classification(self):
        width = height = 10
        multiclass_dataset = FakeDataset([(PIL.Image.new('RGB', (width, height)), [0]), (PIL.Image.new('RGB', (width, height)), 1)])

        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            Task(Task.Config(task_type='classification_multiclass', dirpath=temp_dir)).execute(Task.Inputs(dataset=multiclass_dataset, class_names=["dog", "cat"]))
            with open(temp_dir / 'images.json') as f:
                coco = json.load(f)
                self.assertDictEqual(coco['images'][0], {"id": 1, "width": width, "height": height, "file_name": 'images/1.png'})
                self.assertDictEqual(coco['annotations'][1], {"id": 2, "category_id": 2, "image_id": 2})
                self.assertDictEqual(coco['categories'][1], {"id": 2, "name": "cat"})

        multilabel_dataset = FakeDataset([(PIL.Image.new('RGB', (width, height)), torch.tensor([i, 1-i])) for i in range(2)])
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            json_file_name = 'foo'
            Task(Task.Config(task_type='classification_multilabel', json_file_name=json_file_name, dirpath=temp_dir)).execute(Task.Inputs(dataset=multilabel_dataset, class_names=["dog", "cat"]))
            coco = json.loads((temp_dir / f'{json_file_name}.json').read_text())
            self.assertDictEqual(coco['annotations'][0], {"id": 1, "category_id": 1, "image_id": 1})
            self.assertDictEqual(coco['annotations'][1], {"id": 2, "category_id": 2, "image_id": 1})

    def test_object_detection(self):
        width = height = 10
        dataset = FakeDataset([(torch.zeros((3, height, width)), torch.tensor([[0, 0.1, 0.1, 0.2, 0.2], [1, 0.1, 0.1, 0.2, 0.2]])),
                               (torch.zeros((3, height, width)), [[0, 0.1, 0.1, 0.2, 0.2]]),
                               (torch.zeros((3, height, width)), torch.tensor([]))])
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            image_directory_name = json_file_name = 'foo'

            Task(Task.Config(task_type='object_detection', image_directory_name=image_directory_name, json_file_name=json_file_name, dirpath=temp_dir)).execute(Task.Inputs(dataset, ["dog", "cat"]))
            coco = json.loads((temp_dir / f'{json_file_name}.json').read_text())
            self.assertEqual(len(coco['images']), 3)
            self.assertEqual(len(coco['annotations']), 3)
            self.assertDictEqual(coco['images'][0], {"id": 1, "width": width, "height": height, "file_name": f'{image_directory_name}/1.png'})
            self.assertDictEqual(coco['categories'][1], {"id": 2, "name": "cat"})

            anno = coco['annotations'][1]
            anno['bbox'] = [round(t) for t in anno['bbox']]
            self.assertDictEqual(anno, {"id": 2, "category_id": 2, "image_id": 1, "bbox": [1, 1, 1, 1]})

            anno = coco['annotations'][2]
            anno['bbox'] = [round(t) for t in anno['bbox']]
            self.assertDictEqual(anno, {"id": 3, "category_id": 1, "image_id": 2, "bbox": [1, 1, 1, 1]})



import dataclasses
import typing
import torch
import irisml.core


class DatasetWithTargets(torch.utils.data.Dataset):
    def __init__(self, dataset, targets):
        super().__init__()
        self._dataset = dataset
        self._targets = targets

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, index):
        return self._dataset[index][0], self._targets[index]

    def get_targets(self, index):
        return self._targets[index]


class Task(irisml.core.TaskBase):
    """Make a detection dataset from predictions.

    This task takes a OD dataset and a list of tensor predictions or targets and returns a new dataset
    with the targets set to the provided targets.

    The predictions is a list of tensors, each tensor contains predictions for an image, with shape (N, 5) (targets) or (N, 6) (predictions), each row is class (,score), and box coordinates.

    Output targets bounding boxes will be in xyxy normalized format.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset
        predictions: typing.List[torch.Tensor]

    @dataclasses.dataclass
    class Config:
        normalized: bool = True
        score_threshold: float = 0.

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        predictions: typing.List[torch.Tensor]

    def execute(self, inputs):
        predictions = [torch.as_tensor(p) for p in inputs.predictions]
        predictions = [p[p[:, 1] > self.config.score_threshold] for p in predictions] if predictions[0].shape[1] == 6 else predictions

        if len(inputs.dataset) != len(predictions):
            raise ValueError(f"Dataset and predictions have different lengths: {len(inputs.dataset)} vs {len(predictions)}")

        # Normalized box coordinates.
        if not self.config.normalized:
            for i in range(len(inputs.dataset)):
                if predictions[i].nelement():
                    img = inputs.dataset[i][0]
                    w, h = img.size
                    predictions[i] = predictions[i].float()
                    predictions[i][:, -4] /= w
                    predictions[i][:, -3] /= h
                    predictions[i][:, -2] /= w
                    predictions[i][:, -1] /= h

        predictions = self._correct_coordinates(predictions)
        targets = [torch.cat((p[:, :1], p[:, 2:]), dim=1) for p in predictions] if predictions[0].shape[1] == 6 else predictions

        return self.Outputs(dataset=DatasetWithTargets(inputs.dataset, targets), predictions=predictions)

    def _correct_coordinates(self, targets):
        # Make coordinates within [0, 1].
        return [torch.cat((t[:, :1], torch.clamp(t[:, 1:], min=0., max=1.)), dim=1) for t in targets]

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import torch
from irisml.tasks.find_incorrect_classification_multilabel_indices import Task


class TestFindIncorrectClassificationMultilabelIndices(unittest.TestCase):
    def test_simple(self):
        predictions_list = [torch.tensor([0, 1, 2]), torch.tensor([0, 1, 2])]
        targets_list = [torch.tensor([0, 1, 2]), torch.tensor([0, 1, 2])]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions_list=predictions_list, targets_list=targets_list))
        self.assertEqual(len(outputs.indices), 0)

        predictions_list = [torch.tensor([0, 1, 2]), torch.tensor([0, 1, 2])]
        targets_list = [torch.tensor([0, 1, 2]), torch.tensor([0, 1, 3])]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions_list=predictions_list, targets_list=targets_list))
        self.assertEqual(outputs.indices.tolist(), [1])

        targets = torch.tensor([0, 1])
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions_list=predictions_list, targets=targets))
        self.assertEqual(outputs.indices.tolist(), [0, 1])



import unittest
import unittest.mock
import torch
from irisml.tasks.map_classification_predictions_to_detection import Task


class TestMapClassificationPredictionsToDetection(unittest.TestCase):
    def create_detections_and_index_mappings(self):
        detections = [torch.tensor([[0, 0, 0, 0.5, 0.5], [1, 0, 0, 0.1, 0.1]]), torch.empty(0, 5), torch.tensor([[2, 0, 0, 0.5, 0.5], [3, 0, 0, 0.1, 0.1]])]
        index_mappings = [(0, 0), (0, 1), (2, 0), (2, 1)]
        return detections, index_mappings

    def create_detections_with_score_and_index_mappings(self):
        detections = [torch.tensor([[0, 0.1, 0, 0, 0.5, 0.5], [1, 0.1, 0, 0, 0.1, 0.1]]), torch.empty(0, 5), torch.tensor([[2, 0.8, 0, 0, 0.5, 0.5], [3, 0.8, 0, 0, 0.1, 0.1]])]
        index_mappings = [(0, 0), (0, 1), (2, 0), (2, 1)]
        return detections, index_mappings

    def test_simple(self):
        detections, index_mappings = self.create_detections_with_score_and_index_mappings()
        classifications = torch.tensor([5, 5, 10, 10])
        outputs = Task(Task.Config(score_type='od')).execute(Task.Inputs(classifications, detections, index_mappings))
        self.assertTrue(torch.equal(outputs.detection_predictions[0], torch.tensor([[5, 0.1, 0, 0, 0.5, 0.5], [5, 0.1, 0, 0, 0.1, 0.1]])))
        self.assertTrue(torch.equal(outputs.detection_predictions[1], torch.empty(0, 6)))
        self.assertTrue(torch.equal(outputs.detection_predictions[2], torch.tensor([[10, 0.8, 0, 0, 0.5, 0.5], [10, 0.8, 0, 0, 0.1, 0.1]])))

    def test_ic_probability(self):
        detections, index_mappings = self.create_detections_and_index_mappings()
        classifications = torch.tensor([[0.9, 0.1], [0.8, 0.2], [0.2, 0.8], [0.1, 0.9]])
        outputs = Task(Task.Config(score_type='ic')).execute(Task.Inputs(classifications, detections, index_mappings))
        self.assertTrue(torch.equal(outputs.detection_predictions[0], torch.tensor([[0, 0.9, 0, 0, 0.5, 0.5], [0, 0.8, 0, 0, 0.1, 0.1]])))
        self.assertTrue(torch.equal(outputs.detection_predictions[1], torch.empty(0, 6)))
        self.assertTrue(torch.equal(outputs.detection_predictions[2], torch.tensor([[1, 0.8, 0, 0, 0.5, 0.5], [1, 0.9, 0, 0, 0.1, 0.1]])))

    def test_prod_avg_score(self):
        detections, index_mappings = self.create_detections_with_score_and_index_mappings()
        classifications = torch.tensor([[0.9, 0.1], [0.8, 0.2], [0.2, 0.8], [0.1, 0.9]])
        outputs = Task(Task.Config(score_type='prod')).execute(Task.Inputs(classifications, detections, index_mappings))
        self.assertTrue(torch.allclose(outputs.detection_predictions[0], torch.tensor([[0, 0.09, 0, 0, 0.5, 0.5], [0, 0.08, 0, 0, 0.1, 0.1]])))
        self.assertTrue(torch.equal(outputs.detection_predictions[1], torch.empty(0, 6)))
        self.assertTrue(torch.allclose(outputs.detection_predictions[2], torch.tensor([[1, 0.64, 0, 0, 0.5, 0.5], [1, 0.72, 0, 0, 0.1, 0.1]])))

        outputs = Task(Task.Config(score_type='avg')).execute(Task.Inputs(classifications, detections, index_mappings))
        self.assertTrue(torch.allclose(outputs.detection_predictions[0], torch.tensor([[0, 0.5, 0, 0, 0.5, 0.5], [0, 0.45, 0, 0, 0.1, 0.1]])))
        self.assertTrue(torch.equal(outputs.detection_predictions[1], torch.empty(0, 6)))
        self.assertTrue(torch.allclose(outputs.detection_predictions[2], torch.tensor([[1, 0.8, 0, 0, 0.5, 0.5], [1, 0.85, 0, 0, 0.1, 0.1]])))

    def test_invalid_inputs(self):
        detections, index_mappings = self.create_detections_and_index_mappings()
        index_mappings_shorter = index_mappings[:-1]
        classifications = torch.tensor([[0.9, 0.1], [0.8, 0.2], [0.2, 0.8], [0.1, 0.9]])
        with self.assertRaisesRegex(ValueError, "IC predictions have different samples from the IC to OD index mapping: 4 vs 3"):
            Task(Task.Config()).execute(Task.Inputs(classifications, detections, index_mappings_shorter))

        with self.assertRaisesRegex(ValueError, "When OD scores not provided, score_type has to be ic and IC scores have to be provided!"):
            Task(Task.Config(score_type='od')).execute(Task.Inputs(classifications, detections, index_mappings))

        classifications_no_score = torch.tensor([5, 5, 10, 10])
        detections, index_mappings = self.create_detections_with_score_and_index_mappings()
        with self.assertRaisesRegex(ValueError, "When IC scores not provided, score_type has to be od and OD scores have to be provided!"):
            Task(Task.Config(score_type='ic')).execute(Task.Inputs(classifications_no_score, detections, index_mappings))



import unittest
from irisml.tasks.create_prompt_generator import Task


class TestCreatePromptGenerator(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config(templates=['a photo of a {}'])).execute(Task.Inputs())
        self.assertEqual(outputs.generator('cat'), ['a photo of a cat'])



import dataclasses
import itertools
import logging
import math
import multiprocessing
import os
import typing
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Extracts questions from a VQA dataset.

    __getitem__() of the dataset must return a tuple of ((image, question), answer), where image is a PIL image,
    question is a string, and answer is a string.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        questions: typing.List[str]

    def execute(self, inputs):
        num_processes = len(os.sched_getaffinity(0))
        assert num_processes > 0
        batch_size = math.ceil(len(inputs.dataset) / num_processes)

        logger.info(f"Using {num_processes} processes to collect {len(inputs.dataset)} questions.")

        context = multiprocessing.get_context('fork')
        queue = context.JoinableQueue(num_processes)
        processes = [context.Process(target=Task._worker, args=(inputs.dataset, i, batch_size, queue), daemon=True) for i in range(num_processes)]
        for p in processes:
            p.start()

        collected_batches = [None] * num_processes
        for _ in range(num_processes):
            i, batch = queue.get(timeout=3600)
            collected_batches[i] = batch
            queue.task_done()

        for p in processes:
            p.join()

        assert all(b is not None for b in collected_batches)
        questions = list(itertools.chain.from_iterable(collected_batches))
        logger.info(f"Collected {len(questions)} questions.")
        assert len(questions) == len(inputs.dataset)

        return self.Outputs(questions)

    @staticmethod
    def _worker(dataset, process_index, batch_size, queue):
        start_index = batch_size * process_index
        end_index = min(start_index + batch_size, len(dataset))
        questions = [dataset[i][0][1] for i in range(start_index, end_index)]
        queue.put((process_index, questions))
        queue.join()



"""Compile the model module using torch.compile."""
import logging
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def on_train_start(self, trainer, model):
        logger.info("Compiling the model module.")
        self._original_model = model
        return torch.compile(model)

    def on_train_end(self, trainer, model):
        # The OptimizedModule shares the same state_dict with the original model.
        model = self._original_model
        del self._original_model
        return model

    def on_prediction_start(self, trainer, model):
        return self.on_train_start(trainer, model)

    def on_prediction_end(self, trainer, model):
        return self.on_train_end(trainer, model)
